/*SpringEmbedders - A program to visualize graphs
in non-euclidean geometries using a Springembedder Algorithm

Copyright (C) 2005  Markus Lipp

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#pragma once
#include <math.h>
#include <float.h>

/**PI*/
const float PI       =  3.14159265358979323846f;		// Pi
/**2\*PI*/
const float TWOPI		 =  6.28318530717958647692f;

/**convert degree to radian*/
#define	DEGTORAD(x)	( ((x) * PI) / 180.0 )
/**convert radian to degree*/
#define	RADTODEG(x)	( ((x) * 180.0) / PI )

/** Returns a nonzero value if the float is NAN of not finite*/
#define CHECK_FLOAT(x) ( _isnan(x) | !_finite(x) )
/** Returns a nonzero value if one element of a vector is NAN of not finite*/
#define CHECK_VEC(vec) (  CHECK_FLOAT(vec.x)|CHECK_FLOAT(vec.y)|CHECK_FLOAT(vec.z) )
/** Returns a nonzero value if one element of a vector referenced by a pointer is NAN of not finite*/
#define CHECK_PVEC(vec) (  CHECK_FLOAT(vec->x)|CHECK_FLOAT(vec->y)|CHECK_FLOAT(vec->z) )

/**float machine epsilon*/
const float	EPSILON						= 0.00005f;		
/**Float equality check*/
#define FLOAT_EQ(x,v)	( ((v) - EPSILON) < (x) && (x) < ((v) + EPSILON) )		


/** This is a Basic 3D-Vector class*/
class CVector3
{
public:

		float x,y,z;
		
		CVector3(float sx,float sy,float sz) {x=sx;y=sy;z=sz;};
		CVector3() {x=0;y=0;z=0;};

		void Set(float sx,float sy,float sz) {x=sx;y=sy;z=sz;};

		bool isEqual(CVector3 vec)
		{
			return (FLOAT_EQ(vec.x, x) & FLOAT_EQ(vec.y, y) & FLOAT_EQ(vec.z, z));

		}

		CVector3 RotateY(float r)
		{
			CVector3 res;

			res.Set(cos(r)*x + sin(r)*z,
					y,
					-sin(r)*x + cos(r)*z);

			return res;		

		}

		CVector3 RotateX(float r)
		{
			CVector3 res;

			res.Set(x,
					cos(r)*y-sin(r)*z,
					sin(r)*x+cos(r)*z);

			return res;		

		}

		CVector3 RotateZ(float r)
		{
			CVector3 res;

			res.Set(cos(r)*x-sin(r)*y,
				sin(r)*x+cos(r)*y,
				z);

			return res;		

		}

		void Set(CVector3 vec)
		{
			x=vec.x;
			y=vec.y;
			z=vec.z;
		}
		
		void scale(float s)
		{
			x=x*s;
			y=y*s;
			z=z*s;
		}

		float length()
		{
			return (float)sqrt( (double)pow((double)x,2)+pow((double)y,2)+pow((double)z,2));
		}

		float squaredLength()
		{
			return (float)(pow(x,2)+pow(y,2)+pow(z,2));
		}

		void setLength(float length)
		{
			normalize();
			this->Set( this->mult(length) );

		}

		void normalize() {
			float l = length();
			if (!FLOAT_EQ(l,0.0f))
			{
				x=x/l;y=y/l;z=z/l;
			}
		}
			
		CVector3 cross(CVector3 vec)
		{
			CVector3 out;
			out.Set(
				y*vec.z - vec.y*z,
				z*vec.x - vec.z*x,
				x*vec.y - vec.x*y);
			return out;		
		}

		float* getfArray()
		{
            float* ret = new float[3];
			ret[0]=x; ret[1]=y; ret[2]=z;
            return ret;
		}
		
		CVector3 sub(CVector3 vec) 
		{
			CVector3 out;
			out.Set(x-vec.x,y-vec.y,z-vec.z);
			return out;		
		}

		CVector3 negate() 
		{
			CVector3 out;
			out.Set(-x,-y,-z);
			return out;		
		}
		
		CVector3 add(CVector3 vec) 
		{
			CVector3 out;
			out.Set(x+vec.x,y+vec.y,z+vec.z);
			return out;		
		}
		
		CVector3 add(float sx,float sy,float sz) 
		{
			CVector3 out;
			out.Set(x+sx,y+sy,z+sz);
			return out;		
		}

		bool IsZero()
		{
			if ((x+y+z)==0)
				return true;
			else
				return false;
		}

		void setMax (float m)
		{
			if ( fabs(m) < fabs(x) )
				x = m;
			if ( fabs(m) < fabs(y) )
				x = m;
			if ( fabs(m) < fabs(z) )
				x = m;
		}

		void setMax (CVector3* v)
		{
			if ( fabs(v->x) < fabs(x) )
				x = v->x;
			if ( fabs(v->y) < fabs(y) )
				x = v->y;
			if ( fabs(v->z) < fabs(z) )
				x = v->z;
		}

		CVector3 mult(float s) 
		{
			CVector3 out;
			out.Set(x*s,y*s,z*s);
			return out;		
		}

		CVector3 mult(CVector3 s) 
		{
			CVector3 out;
			out.Set(x*s.x,y*s.y,z*s.z);
			return out;		
		}

		CVector3 div(CVector3 s) 
		{
			CVector3 out;
			out.Set(x/s.x,y/s.y,z/s.z);
			return out;		
		}

		CVector3 abs() 
		{
			CVector3 out;
			out.Set( fabs(x),fabs(y),fabs(z));
			return out;		
		}

		CVector3 div(float s) 
		{
			CVector3 out;
			out.Set(x/s,y/s,z/s);
			return out;		
		}
		
		float dot(CVector3 vec)
		{
			return vec.x*x+vec.y*y+vec.z*z;
		}

		
		void RotateArbitrary(float angle, float ax, float ay, float az)
		{
			float cosTheta = cos(angle);
			float sinTheta = sin(angle);
			float xn,yn,zn;

		
			xn  = (cosTheta + (1 - cosTheta) * ax * ax)		* x;
			xn += ((1 - cosTheta) * ax * ay - az * sinTheta)	* y;
			xn += ((1 - cosTheta) * ax * az + ay * sinTheta)	* z;


			yn  = ((1 - cosTheta) * ax * ay + az * sinTheta)	* x;
			yn += (cosTheta + (1 - cosTheta) * ay * ay)		* y;
			yn += ((1 - cosTheta) * ay * az - ax * sinTheta)	* z;

		
			zn  = ((1 - cosTheta) * ax * az - ay * sinTheta)	* x;
			zn += ((1 - cosTheta) * ay * az + ax * sinTheta)	* y;
			zn += (cosTheta + (1 - cosTheta) * az * az)		* z;

			x=xn;
			y=yn;
			z=zn;

		}


		void TransformByMatrix(float* matrix)
		{
			float x2,y2,z2;

            x2 = x*matrix[0]+y*matrix[4]+z*matrix[8]+matrix[12];
			y2 = x*matrix[1]+y*matrix[5]+z*matrix[9]+matrix[13];
			z2 = x*matrix[2]+y*matrix[6]+z*matrix[10]+matrix[14];

			x=x2; y=y2; z=z2;
		}

		CVector3 LinearAttenduationComponents(CVector3 vec2, float max, float k)
		{
            CVector3 out;
			out = sub(vec2);
			out = out.abs();
			out = out.mult(k);
			out.setMax(max);
			return out;
		}

		CVector3 CalcOrthogonalAngles()
		{
            CVector3 out,xv(1,0,0),yv(0,1,0),zv(0,0,1),h;
			
			//Take each Component and calculate angle
			h.Set(x,y,0);
			h.normalize();
            out.x = acos(h.dot(xv));			

			h.Set(x,y,0);
			h.normalize();
            out.y = acos(h.dot(yv));			

			h.Set(0,y,z);
			h.normalize();
            out.x = acos(h.dot(zv));			

			return out;
		}		

		float LinearAttenduationLength(CVector3 vec2, float max, float k)
		{
            float out;
			out = (sub(vec2)).length();
			out = fabs(out)*k;			
			if (out > max)
				out=max;
			return out;
		}

		CVector3 LinearInterpolate(CVector3 vec2, float k)
		{
			CVector3 out;
			out = sub(vec2);
			out.scale(-k);
			return out.add(*this);
		}




		void TransformByMatrix3(float* matrix)
		{
			float x2,y2,z2;

            x2 = x*matrix[0]+y*matrix[4]+z*matrix[8];
			y2 = x*matrix[1]+y*matrix[5]+z*matrix[9];
			z2 = x*matrix[2]+y*matrix[6]+z*matrix[10];
			x=x2; y=y2; z=z2;
		}



		void TransformByMatrixTranspose(float* matrix)
		{
			float x2,y2,z2;

			x = x-matrix[12];
			y = y-matrix[13];
			z = z-matrix[14];

            x2 = x*matrix[0]+y*matrix[1]+z*matrix[2];
			y2 = x*matrix[4]+y*matrix[5]+z*matrix[6];
			z2 = x*matrix[8]+y*matrix[9]+z*matrix[10];
			
			x=x2; y=y2; z=z2;
		}

    

		void TransformByMatrix3Transpose(float* matrix)
		{
			float x2,y2,z2;

            x2 = x*matrix[0]+y*matrix[1]+z*matrix[2];
			y2 = x*matrix[4]+y*matrix[5]+z*matrix[6];
			z2 = x*matrix[8]+y*matrix[9]+z*matrix[10];
			x=x2; y=y2; z=z2;
		}

		void TransformByMatrix4Transpose(float* matrix)
		{
			float x2,y2,z2;

            x2 = x*matrix[0]+y*matrix[1]+z*matrix[2]-matrix[3];
			y2 = x*matrix[4]+y*matrix[5]+z*matrix[6]-matrix[7];
			z2 = x*matrix[8]+y*matrix[9]+z*matrix[10]-matrix[11];
			x=x2; y=y2; z=z2;
		}

		void  setRandom(float max, float offset)
		{

            x = ((float)rand()/(float)RAND_MAX)*max +offset;
			y = ((float)rand()/(float)RAND_MAX)*max +offset;
			z = ((float)rand()/(float)RAND_MAX)*max +offset;
		}

			
		
};

