/*SpringEmbedders - A program to visualize graphs
in non-euclidean geometries using a Springembedder Algorithm

Copyright (C) 2005  Markus Lipp

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/* The non-Euclidic spring-Algorithm used in this program was proposed by:

Stephen G. Kobourov und Kevin Wampler, Non-Euclidean Spring Embedders, in INFOVIS 2004 pages 207-214

*/


#pragma once

#include "Vector3.h"
#include <math.h>
#include <complex>




/**
*Base class for all geometry types. Defines all functions necessary to use 
*any geometry with the spring-embedder algorithm.
*/
class cGeometry
{
public:

	/** Transform vector y to Tangent Space from vector x*/
	virtual CVector3 transformTangent(CVector3 x, CVector3 y) {return y.sub(x);};
	/**Inverse-Transform of tangentspace-vector txy.*/
	virtual CVector3 transformTangentInverse(CVector3 x, CVector3 txy) {return txy.add(x);}
	/**Set up all one-time constants needed for further calculations 
	in tangentspace for vector x*/
	virtual void setupTangentSpace(CVector3 x) {};
	

	/**Interpolate and draw geodesic from vec1 to vec2, using "steps" linear elements at maximum*/
	virtual void interpolateDraw(CVector3 vec1, CVector3 vec2, int steps) {}

	/**Return max coordinate position, used for randomizing Positions of this geometry */
	virtual float getMaxPosition() {return 0;};

	cGeometry(void)
	{}
	~cGeometry(void)
	{}

	/**Draw axis of this geometry*/
	virtual 	void DrawAxis() {};

	/**Transform vec to 3D-Euclidic Space*/
	virtual CVector3 ProjectToEuclid(CVector3 vec) {return vec;};

	/**Transform vec from 3D-Euclidic Space to geometry space*/
	virtual CVector3 ProjectFromEuclid(CVector3 vec) {return vec;};

	/**Is a simple line enough for interpolation?*/
	virtual bool isLinearInterpolable() {return true;};

	/**return vec modulo to geometry boundaries.*/
	virtual CVector3 modulo(CVector3 vec)
	{
		return vec;
	};
};


/**
*Implements an Euclidian 3D Geometry
*/
class cEuclidianGeometry:cGeometry 
{
public:


	void DrawAxis()	
	{
		
	};

	float getMaxPosition()	{return 50;};

	CVector3 ProjectToEuclid(CVector3 vec) {return vec;};

	bool isLinearInterpolable() {return true;};
	CVector3 modulo(CVector3 vec) {return vec;};

};

/**
*Implements a Spherical 2D Geometry
*/
class cSphericalGeometry:cGeometry 
{
public:

	void DrawAxis()
	{
		//glDepthMask(false); 
		glColor3f(0.7,0.7,0.7);

		glEnable(GL_BLEND);
		glEnable(GL_CULL_FACE);
		glBlendFunc(GL_ONE,GL_ONE);

		gluSphere(m_glqSphere, m_fRadius-0.2,32,32);
		glDisable(GL_BLEND);
		glDisable(GL_CULL_FACE);
		//glDepthMask(true); 

	}

	cSphericalGeometry()
	{
		m_glqSphere = gluNewQuadric(  );
        m_fRadius = 1;
	}

	void setupTangentSpace(CVector3 x)
	{
		//Formulas from source paper:
		//u = (-sin theta sin phi, 0, cos theta sin phi)
		//v = (cos theta cos phi, - sin phi, sin theta cos phi).

		float phi = x.x;
		float theta = x.y; 

		if ( FLOAT_EQ(phi,0)  )
		{
			//We are at the pole of sphere
			//Use fixed uv-set
			m_vecU.Set(1,0,0);
			m_vecV.Set(0,0,-1);

		}
		else
		{
			if ( FLOAT_EQ(phi,PI) )
			{
				//We are at the other pole of sphere
				//Use fixed uv-set
				m_vecU.Set(1,0,0);
				m_vecV.Set(0,0,1);
			}
			else
			{
                //Calculate tangent vectors
				m_vecU.Set(-sin(theta)*sin(phi), 0, cos(theta)*sin(phi));
				m_vecV.Set(cos(theta)*cos(phi),-sin(phi), sin(theta)*cos(phi));		
			}
		}

		//Normalize tangent vectors
		m_vecU.normalize();
		m_vecV.normalize();
		
		//Orthonormalize tangent basis
		CVector3 up = m_vecV.cross(m_vecU);
		CVector3 un = up.cross(m_vecV);
		CVector3 vn = up.cross(m_vecU);

		m_vecU.Set(un);
		m_vecV.Set(vn);

        //Calculate euclidic projection of vector x, needed for further calculations
		m_vecXProjected = ProjectToEuclid(x);

		float phix = x.x;	
		float thetax = x.y; 

		//these values are used later on
		m_fSinphix = sin(phix);
		m_fCosphix = cos(phix);
	}

	CVector3 transformTangent(CVector3 x, CVector3 y)
	{

		float phiy = y.x;
		float thetay = y.y; 
		float phix = x.x;	
		float thetax = x.y; 
		CVector3 Projection;			

		//Calculate embedding of y in euclidic space
		CVector3 EmbeddingY = ProjectToEuclid(y);
		//calculate relative embedding to x
		EmbeddingY = EmbeddingY.sub(m_vecXProjected);

		//calculate projection to tangent space - length is not yet correct!
		Projection = CVector3(EmbeddingY.dot(m_vecU),EmbeddingY.dot(m_vecV),0);

		//calculate gedoesic length (length of shortest path) between x and y
		//cos of length:
		float t=m_fSinphix*sin(phiy)*cos(thetay - thetax) + m_fCosphix*cos(phiy) ;

		//clamp cos-value to avoid error with acos
		t = __max(__min(t,1),-1);

		//set projection to correct length
		Projection.setLength( m_fRadius*acos(t) );//
		
		return Projection;	
	}

	CVector3 transformTangentInverse(CVector3 x,CVector3 txy)
	{
		CVector3 result;

		//Rotate x around rotation axis with specific angle

		//Calculate rotation axis
		
		//Rotate txy 90 degrees -> temp
		CVector3 temp (-txy.y, txy.x, 0); 
		temp.normalize();
		//temp is not perpendicular to txy

		//calculate relative euclidic position of temp from x 
		CVector3 axis =  (m_vecU.mult(temp.x)).add(m_vecV.mult(temp.y));

		axis.normalize();

		//Calculate rotation angle
		float angle = txy.length() / m_fRadius;

		//Project x to euclidic 3D space
		result = ProjectToEuclid(x);

		//Rotate 
		result.RotateArbitrary(-angle, axis.x, axis.y, axis.z);		

		//Reproject
		result = ProjectFromEuclid(result);

		return result;
	};


	void interpolateDraw(CVector3 vec1, CVector3 vec2, int steps)
	{

		CVector3 result1, result2;

		//calculate angle between input vectors
		CVector3 q0 = vec1; 
		CVector3 q1 = vec2; 
		q0.normalize();
		q1.normalize();
		float dot = q0.dot(q1);
		dot = __max(__min(dot,1),-1);
		float phi = acos(dot);  

		//adaptive stepsize:
		//Use maximum if angle phi=PI
		steps = (int)ceil(steps*phi/PI);
	
		//formulae used for slerp (spherical linear interpolation):
		//slerp( t, q0, q1 ) = [ q0*sin(phi*(1-t)) + q1*sin(phi*t ) ] / sin(phi)

		float dist = 0;
		result1 = q0;

		glBegin(GL_LINES);
		for (int j=0; j<steps;j++)
		{
			glVertex3f(result1.x*m_fRadius, result1.y*m_fRadius, result1.z*m_fRadius);
			float dist = ((float)j+1.0)/(float)steps;
			result2 = (q0.mult( sin(phi*(1-dist)) ).add(q1.mult(sin(phi*dist)))).div(sin(phi));
			glVertex3f(result2.x*m_fRadius, result2.y*m_fRadius, result2.z*m_fRadius);
			result1=result2;


		}
		glEnd();
	
	
	}

	CVector3 ProjectToEuclid(CVector3 vec)
	{
		float theta=vec.y;
		float phi = vec.x;
		//Formulae from source paper:
		CVector3 result =  CVector3(cos(theta)*sin(phi)*m_fRadius, cos(phi)*m_fRadius, sin(theta)*sin(phi)*m_fRadius);


		return result;

	};

	CVector3 ProjectFromEuclid(CVector3 vec)
	{

		CVector3 result;
		
		vec.normalize();

		//Calculate angle theta
		float at = atan2(vec.z,vec.x);
		//at is now in range (-PI,PI)
		//we now move it to (0,2PI)
		if (at<0)
			at = PI + (PI+at);

		//fetch cos(phi), clamp to [-1,1]
		float y = __max(__min(vec.y,1),-1);
			
		//combine results
		result = CVector3(acos(y),(at),0);

		return result;
	};

	float getMaxPosition()	{return PI;};

	/** set Radius of Sphere*/
	void setRadius(float r)
	{
		m_fRadius=r;
	}

	float getRadius() {return m_fRadius;}

	bool isLinearInterpolable() {return false;};

	CVector3 modulo(CVector3 vec)
	{
		float theta = vec.y;
		float phi = vec.x;

		//phi should be in range [0,PI], theta in [0,2PI]

		if  (phi > PI)
		{
			//theta is dependent on phi if phi>PI
			theta+=PI;
			phi = TWOPI - phi;
		}

		if (phi < 0)
		{
			theta+=PI;
			phi = - phi;
		}

		//calculate modulo of theta
		theta = fmod (theta, TWOPI);

		return CVector3(phi,theta,0);
	};

private:

	//this variables store information needed for tangent-space transformation
	//that do not change for different vertices to be transformed
	CVector3 m_vecU,m_vecV;
	CVector3	m_vecXProjected ;
	float	m_fSinphix , m_fCosphix;


	float m_fRadius;
	GLUquadricObj* m_glqSphere;

};

/**
*Implements a Hyperbolic 2D Geometry
*/

/// @cond
using namespace std;
 /// @endcond

class cHyperbolicGeometry:cGeometry 
{
	

public:


	cHyperbolicGeometry()
	{
		m_fPoincareRadius=19;
		m_complexCenterOfAttention = complex <double> (0.0,0.0);
	}

	CVector3 transformTangent(CVector3 x, CVector3 y)	
	{
		using namespace std;


		complex <double> z0 ( (double)x.x , (double)x.y );
		complex <double> z ( (double)y.x , (double)y.y );
		complex <double> result;

		//Formulae from source paper:
		//transform x to center at first
		result = (z-z0)/(1.0-(conj(z0)*z));
		double absresult = abs(result);

		if (FLOAT_EQ(absresult,0.0))
		{
			
		}
		else
		{
			//only if magnitude of transformed y is not zero:
			//calculate hyperbolic length of y
			double l = __max(0.0001, (1.0+absresult)/(1.0-absresult));
			result = (result/absresult)*log(l);
		}
		
		//set correct tangentspace vector
		CVector3 ret = CVector3((float)real(result), (float)imag(result),0.0);
		
		return ret;
	
	};


	CVector3 transformTangentInverse(CVector3 x, CVector3 txy) 
	{
		using namespace std;


		complex <double> z0 ((double) x.x , (double)x.y );
		complex <double> z ( (double)txy.x , (double)txy.y );
		complex <double> result;

		double absz = abs(z);
		//as we use exp(): clamp absz to [0,700]
		absz=min(absz, 700.0);

		if (FLOAT_EQ(absz,0.0))
		{
			//result=txy
			result=z;            
		}
		else 
		{
			//Calculate length on poincare disc from hyperbolic length
			//from tangentspace
			result = (z/absz)*abs((1.0-exp(absz))/(1.0+exp(absz)));     
		}

		//move x result back from center
		result = (-result-z0)/( (-(conj(z0))*result)-1.0);

		CVector3 ret = CVector3((float)real(result), (float)imag(result),0.0);
		
		return ret;
	}

	

	//virtual CVector3 interpolate(CVector3 vec1, CVector3 vec2, float dist) {return vec1;}

	
	float getMaxPosition() {return 0.7f;};

	CVector3 ProjectToEuclid(CVector3 vec)	
	{
		using namespace std;

		complex <double> z ( (double)vec.x , (double)vec.y );
		complex <double> result;

		//transform m_complexCenterOfAttention to center
		result = (z-m_complexCenterOfAttention)/(1.0-(conj(m_complexCenterOfAttention)*z));

        //multiply transformed vec with m_fPoincareRadius
		return CVector3(real(result), imag(result),0).mult(m_fPoincareRadius);
	
	};

	CVector3 ProjectFromEuclid(CVector3 vec)	
	{
		using namespace std;

		complex <double> z ( (double)vec.x , (double)vec.y );
		complex <double> result;
		
		//transform m_complexCenterOfAttention back from center
		result = (-z-m_complexCenterOfAttention)/( (-(conj(m_complexCenterOfAttention))*z)-1.0);

		//divide transformed vec with m_fPoincareRadius
		return CVector3(real(result), imag(result),0).div(m_fPoincareRadius);
	
	};


	CVector3 ProjectFromEuclidInverseCOA(CVector3 vec)	
	{
		//needed to calculate position of mousecursor relative to actual center of attention

		using namespace std;

		//at first scale to length of vec [-1,1]
		vec = vec.div(m_fPoincareRadius);

        complex <double> z ( (double)vec.x , (double)vec.y );
		complex <double> result;

		//transform m_complexCenterOfAttention back from center
		result = (-z-m_complexCenterOfAttention)/( (-(conj(m_complexCenterOfAttention))*z)-1.0);

		//we now have the poincare position of the euclidic position with respect to the center of attention
		return CVector3(real(result), imag(result),0);
	
	};

	bool isLinearInterpolable()
	{
		return false;
	};

	CVector3 modulo(CVector3 vec)
	{
		//Just clamp to 0.99 to avoid overflows
		if (vec.length() > 0.99)
			vec.setLength(0.99);

		return vec;
	};

	/** Set radius of poincare-disc*/
	void setPoincareRadius(float r)
	{
		m_fPoincareRadius=r;
	}

	
	float getPoincareRadius()
	{
		return m_fPoincareRadius;
	}

	void interpolateDraw(CVector3 vec1, CVector3 vec2, int steps)
	{
		

		//Calculate angle between vec1 and vec2
		//To test if they lie on a line
		vec1 = vec1.div(m_fPoincareRadius);
		vec2 = vec2.div(m_fPoincareRadius);

		CVector3 vec1t = vec1;
		CVector3 vec2t = vec2;

		vec1t.normalize();
		vec2t.normalize();

		float dot = vec1t.dot(vec2t);
		dot = __max(__min(dot,1),-1);
		float angle = acos(dot);

		float tol = 0.01;

		if ( ( ((angle) - tol) < (PI) && (PI) < ((angle) + tol) )
			| ( ((angle) - tol) < (0) && (0) < ((angle) + tol) ))
		{
			//They lie on a line

			glBegin(GL_LINES);
			glVertex3f(vec1.x*m_fPoincareRadius, vec1.y*m_fPoincareRadius, vec1.z*m_fPoincareRadius);
			glVertex3f(vec2.x*m_fPoincareRadius, vec2.y*m_fPoincareRadius, vec2.z*m_fPoincareRadius);
			glEnd();

		}
		else
		{

			//Calculate Midpoint of circle containing vec1 and vec2
			//Formulae is derived here
			//http://www.ms.uky.edu/~carl/ma502/html/hyperb2.html by Carl Eberhart
			

			float a=vec1.x;
			float b=vec1.y;
			float c=vec2.x;
			float d=vec2.y;

			CVector3 vCenter
				( (0.5*d*a*a-0.5*b*c*c+0.5*d*b*b-0.5*b*d*d-0.5*b+0.5*d)/(d*a-b*c), 
				(-0.5*c*b*b+0.5*d*d*a-0.5*c*a*a+0.5*c*c*a+0.5*a-0.5*c)/(d*a-b*c),
				0);

			
			float radius = (vCenter.sub(vec1)).length();

			//Now that we have radius and center of circle, 
			//we just do a slerp between vec1 and vec2

			CVector3 result1, result2;

			CVector3 q0 = vec1.sub(vCenter);
			q0.normalize();	
			CVector3 q1 = vec2.sub(vCenter);
			q1.normalize();		

			//calculate angle			
			float dot = q0.dot(q1);
			dot = __max(__min(dot,1),-1);
			float phi = acos(dot);  
		
			//slerp( t, q0, q1 ) = [ q0*sin(phi*(1-t)) + q1*sin(phi*t ) ] / sin(phi)

			float dist = 0;
			result1 = vec1.mult(m_fPoincareRadius);

			//adaptive stepsize:
			//calculate length of arc, use max steps for arclength==2.0
			//as arc can not get longer in poincare disc
			steps = (int)ceil((float)steps* (phi*radius)/2.0);

			glBegin(GL_LINES);

			for (int j=0; j<steps;j++)
			{
				glVertex3f(result1.x, result1.y, result1.z);

				float dist = ((float)j+1.0)/(float)steps;
				result2 = (q0.mult( sin(phi*(1-dist)) ).add(q1.mult(sin(phi*dist)))).div(sin(phi));
				result2 = ((result2.mult(radius)).add(vCenter)).mult(m_fPoincareRadius);
				glVertex3f(result2.x, result2.y, result2.z);
				result1=result2;


			}
	
			glEnd();
		}
		
	}

	void DrawAxis()	
	{
		CVector3 result1(m_fPoincareRadius,0,0);
		CVector3 result2;

		//Draw poincare disc
		glBegin(GL_LINES);
		for (int j=0; j<50;j++)
		{
			glVertex3f(result1.x, result1.y, result1.z);

			float dist = ((float)j+1.0)/(float)50;
			float alphaT = 2*PI*dist;
			result2 = CVector3( m_fPoincareRadius*cos(alphaT) ,m_fPoincareRadius*sin(alphaT) ,0.0f);
			glVertex3f(result2.x, result2.y, result2.z);
			result1=result2;

		}
		glEnd();
	}

	/** Set the center of attention on the poincare disc*/
	void setCenterOfAttention(complex <double> coa)
	{
		m_complexCenterOfAttention=coa;
	}

	/** Set the desired center of attention on the poincare disc.*/
	void setDesiredCenterOfAttention(complex <double> coa)
	{
		complex <double> result;
		
		//transform m_complexCenterOfAttention from center
		result = (-coa-m_complexCenterOfAttention)/( (-(conj(m_complexCenterOfAttention))*coa)-1.0);

		//calculate position of desired coa
		CVector3 temp(real(result),imag(result),0);
		if (temp.length() > 0.99)
			temp.setLength(0.99);

		m_complexDesiredCenterOfAttention=complex <double> (temp.x, temp.y);
	}

	/** Set the desired center of attention on the poincare disc. Dont transform m_complexCenterOfAttention from center at first*/
	void setDesiredCenterOfAttentionDirect(complex <double> coa)
	{
		//Dont transform m_complexCenterOfAttention from center at first
		m_complexDesiredCenterOfAttention=coa;
	}

	complex <double> getDesiredCenterOfAttention()
	{
		return m_complexDesiredCenterOfAttention;
	}

	/**force directed movement of current center of attention to the desired center of attention */
	void IntegrateCenterOfAttention(float timestep)
	{
		

		//setup vectors
        CVector3 dcoa (real(m_complexDesiredCenterOfAttention), imag(m_complexDesiredCenterOfAttention),0 );
		CVector3 coa (real(m_complexCenterOfAttention), imag(m_complexCenterOfAttention),0 );

		//calculate distance between current coa to desired coa
        CVector3 F = dcoa.sub(coa);

		//Move proportional to force and timestep
        coa = coa.add(F.mult(10*timestep));

		//clamp to 0.99 to avoid overflow
		if (coa.length() > 0.99)
			coa.setLength(0.99);

		//set new coa
		m_complexCenterOfAttention = complex <double> (coa.x, coa.y);
	}


private:

	float m_fPoincareRadius;
	complex <double> m_complexCenterOfAttention,m_complexDesiredCenterOfAttention;

};

/**Handle mathematical exceptions*/
int _matherr( struct _exception *except )
{
    except->retval=0;

    if( except->type == _DOMAIN )
    {		
		//MessageBox(NULL, except->name,"Argument domain error", MB_OK);
		return 1;
	}
    if( except->type == _SING )
    {		
		//MessageBox(NULL, except->name,"Argument singularity", MB_OK);
		return 1;
	}
    if( except->type == _OVERFLOW )
    {		
		//MessageBox(NULL, except->name,"Overflow range error", MB_OK);
		return 1;
	}
    if( except->type == _PLOSS )
    {		
		//MessageBox(NULL, except->name,"Partial loss of significance", MB_OK);
		return 1;
	}
    if( except->type == _PLOSS )
    {		
		//MessageBox(NULL, except->name,"Total loss of significance.", MB_OK);
		return 1;
	}
    if( except->type == _UNDERFLOW	 )
    {		
		//MessageBox(NULL, except->name,"Underflow", MB_OK);
		return 1;
	}

	   return 0;    
}