/*SpringEmbedders - A program to visualize graphs
in non-euclidean geometries using a Springembedder Algorithm

Copyright (C) 2005  Markus Lipp

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/* The non-Euclidic spring-Algorithm used in this program was proposed by:

Stephen G. Kobourov and Kevin Wampler, Non-Euclidean Spring Embedders, in INFOVIS 2004 pages 207-214

*/

#pragma once

#include "Vector3.h"
#include "cSpring.h"
#include "cGeometry.h"
#include <stdlib.h>
#include <stdio.h>
#include <fstream>
#include <sys\stat.h>
#include <iostream>
#include <stdio.h>
#include <string.h>
#include <xercesc/dom/DOM.hpp>
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/XMLString.hpp>
#include <iostream>
#include <xercesc/framework/StdOutFormatTarget.hpp>
#include <xercesc/dom/impl/DOMWriterImpl.hpp>
#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/framework/LocalFileInputSource.hpp>>
#include "DOMTreeErrorReporter.hpp"

#ifdef XERCES_CPP_NAMESPACE_USE
XERCES_CPP_NAMESPACE_USE
#endif

/**Random number in Range [0,1]*/
#define RAND01 (float)rand()/(float)RAND_MAX
/**Random number in Range [0,x]*/
#define RANDX(x) RAND01*x
/**Random number in Range [-x,x]*/
#define RANDXU(x)  (RAND01-0.5f)*x*2.0f
/** result is one with probability of 0.5*/
#define RANDB	(( rand()> (RAND_MAX/2)) ? true:false )
/** result is one with probability of x*/
#define RANDBP(x)	(( rand() > ((float)RAND_MAX*x)) ? false:true )
/** Transcode between char and XMLchar*/
#define XML_TC(x) XMLString::transcode(x)



/**This class encapsulates everything related to graph loading and managing.
The springembedder-algorithm is implemented in this class too.
*/
class cGraph
{
public:


	/** Load graph from xml file */
	void loadXMLGraph(char* filename, XercesDOMParser* parser, DOMTreeErrorReporter* errHandler)
	{
		m_fMaxScale=1;

		//open file&parse file
		bool errorsOccured=false;
	    try
		{
			parser->parse(filename);

		}

		//exception handlers
		catch (const XMLException& e)
		{
			XERCES_STD_QUALIFIER cerr << "An error occurred during parsing\n   Message: "
				<< StrX(e.getMessage()) << XERCES_STD_QUALIFIER endl;
			errorsOccured = true;
		}

		catch (const DOMException& e)
		{
			const unsigned int maxChars = 2047;
			XMLCh errText[maxChars + 1];

			XERCES_STD_QUALIFIER cerr << "\nDOM Error during parsing: '" << filename << "'\n"
				<< "DOMException code is:  " << e.code << XERCES_STD_QUALIFIER endl;

			if (DOMImplementation::loadDOMExceptionMsg(e.code, errText, maxChars))
				XERCES_STD_QUALIFIER cerr << "Message is: " << StrX(errText) << XERCES_STD_QUALIFIER endl;

			errorsOccured = true;
		}

		catch (...)
		{
			XERCES_STD_QUALIFIER cerr << "An error occurred during parsing\n " << XERCES_STD_QUALIFIER endl;
			errorsOccured = true;
		}

		bool t = errHandler->getSawErrors();

		if (t==false)
		{
			//continue if no errors occurred

			//get DOM document
			m_pDoc = parser->getDocument();
			//extract root node
			DOMElement* rootElement = m_pDoc->getDocumentElement();
			//get list of children
			DOMNodeList* children = rootElement->getChildNodes();
			
			//variables needed while stepping trough the DOM
			DOMNode *node, *attributenode;
			DOMNamedNodeMap* attributes;
			string all;

			//Count number of vertices
			int numVertices=0;
			for (int i=0; i<children->getLength();i++)
			{
				if ( strcmp(XML_TC((children->item(i))->getNodeName()),"node" )==0)
					numVertices++;
			}			
			
			//Initialize graph data structures based on numVertices
			Init(numVertices);

            //Extract id, label and weights	of vertices
			numVertices=0;

			for (int i=0; i<children->getLength();i++)
			{
				node = children->item(i);
				//get attributes
				attributes = node->getAttributes();

				if ( strcmp(XML_TC(node->getNodeName()),"node" )==0)
				{
					//if this is a node:
					
					//get id attribute node
					attributenode = attributes->getNamedItem(XML_TC("id"));				
					//fetch id attribute
					m_psVerticeNames[numVertices] = string(XML_TC(attributenode->getTextContent()));

					//get label attribute node
					attributenode = attributes->getNamedItem(XML_TC("label"));				
					//fetch label attribute
					m_psVerticeLabels[numVertices] = string(XML_TC(attributenode->getTextContent()));

					attributenode = attributes->getNamedItem(XML_TC("weight"));				
					//get weight attribute node
					m_psVerticeWeights[numVertices] = atof( string(XML_TC(attributenode->getTextContent())).data());

					//update maximum encountered weight
					//used to scale proportional while rendering
					if (m_psVerticeWeights[numVertices]>m_fMaxScale)
						m_fMaxScale=m_psVerticeWeights[numVertices];

					//No weight? -> set to 1
					if (FLOAT_EQ(m_psVerticeWeights[numVertices],0))
						m_psVerticeWeights[numVertices]=1;

					numVertices++;
				}
			}

			
			//go trough all edges
			for (int i=0; i<children->getLength();i++)
			{
				node = children->item(i);
				attributes = node->getAttributes();

				if ( strcmp(XML_TC(node->getNodeName()),"edge" )==0)
				{
					//get attribute nodes to target&source
					//fetch source/target name
					attributenode = attributes->getNamedItem(XML_TC("source"));
					string ref1 = string(XML_TC(attributenode->getTextContent()));
					attributenode = attributes->getNamedItem(XML_TC("target"));
					string ref2 = string(XML_TC(attributenode->getTextContent()));

					int ref1_node=0;
					int ref2_node=0;

					//search in vertices array for those names
					for (int i=0; i< getNumNodes(); i++)
					{
						if ( ref1==m_psVerticeNames[i] )
							ref1_node=i;
						if ( ref2== m_psVerticeNames[i] )
							ref2_node=i;                        
					}

                    //set connection
					setIsConnected(ref1_node, ref2_node, true);
				}
			}

		}
	
		
		ClearvecForces();
		ClearvecVelocities();
		ClearIsSelected();
	}
	

    /**Just randomize a graph*/
	void LoadGraph()
	{
		Init(20);

		for (int i=0; i<getNumNodes(); i++)
			for (int ii=i; ii<getNumNodes(); ii++)
			{
				setIsConnected(i,ii,RANDBP(0.03f));
               
			}

		ClearvecForces();
		ClearvecVelocities();
		ClearIsSelected();
	}

	/**Update node-positions with springembedders algorithm*/
	void SpringAlgorithm(cGeometry* geom, cSpring* spring, float timestep, float mass)
	{
		//scale timestep
        timestep=m_fTimeStepMult/1000.0f;

		//Set Forces to zero
		ClearvecForces();
		CVector3 Position1, Position2,Position2_Tangent, Force;

		//Calculate Forces
        //Step trough all nodes
		for (int i=0; i<getNumNodes(); i++)
		{
			Position1 = getPosition(i);
			//Setup tangentspace for current node
			geom->setupTangentSpace(Position1);

			for (int ii=0; ii<getNumNodes(); ii++)
			{
				//skip for same node
				if (i==ii)
					continue;


				Position2 = getPosition(ii);

				//transform node ii to tangentspace from node i

				Position2_Tangent = geom->transformTangent(Position1, Position2);

				//Calculate repulsion force
				Force = spring->forceRepulsion(Position2_Tangent);

				//add force, only if Force is well-formed (i.e. not infinite)
				if (!CHECK_VEC(Force))
					AddForce(i, Force);
				
				if (isConnected(i,ii))
				{
					//calculate spring connection force if connected
					Force = spring->forceConnected(Position2_Tangent);
					if (!CHECK_VEC(Force))
						AddForce(i, Force.negate());					
				}
               
			}


			CVector3 AccelerationTangent;

			//Move proportional to force
			
			//F=m*a => a = F/m
			AccelerationTangent = m_pvecForces[i].div(mass);

			//we do not integrate velocities over time!!
			//this means that no inertia is simulated!
			//it was observed that this algorithm works better without inertia.
			//(for physical plausible simulations we should integrate velocities)
			
			//a = dv/dt => dv = a*dt
			SetVelocity(i, AccelerationTangent.mult(timestep*10));

			//calculate tangent-space position of moved node
			//ds=v*dt
			CVector3 newPointTangent =  m_pvecVelocities[i].mult(timestep);	

			//project back to geometry space
            CVector3 newpos = geom->transformTangentInverse(Position1, newPointTangent);

			//only set new position if it is not malformed
			if (!CHECK_VEC(newpos))
				SetPosition(i, newpos, geom);	


		}

	
		


	}

	string getNodeLabel(int Node)
	{
		if (Node<m_uiNumNodes)
			return m_psVerticeLabels[Node] ;
		else 
			return string();
	}


	unsigned int getNumNodes()
	{
		return m_uiNumNodes;
	}


	/**Randomize all node Positions. Maximum value is determined by geometry.*/
	void RandomizePositions(cGeometry* geom)
	{		
		float max = geom->getMaxPosition();
	
		for (int i=0; i<m_uiNumNodes; i++)
		{
			if (!is2DRepresentation())
				SetPosition(i, CVector3(RANDXU(max),RANDXU(max),RANDXU(max)),geom );
			else
				SetPosition(i, CVector3(RANDXU(max),RANDXU(max),0),geom );

		}
		ClearvecForces();
		ClearvecVelocities();
		ClearIsSelected();
	}

	/**Get Position from current position-set*/
	CVector3 getPosition(unsigned int Node)
	{
		if (Node<m_uiNumNodes)
			switch (m_iUsedSet)
			{
				case 1: return m_pvecPositions1[Node] ;break;
				case 2: return m_pvecPositions2[Node] ;break;
				case 3: return m_pvecPositions3[Node] ;break;
			}
		else 
			return CVector3(0,0,0);

	}

	float getWeight(unsigned int Node)
	{
		if (Node<m_uiNumNodes)
			return m_psVerticeWeights[Node] ;
		else 
			return 0;

	}

	/**set position in current position set*/
	void SetPosition(unsigned int Node, CVector3 Position, cGeometry* geom)
	{

		if (Node<m_uiNumNodes)
		{
			CVector3 pos = geom->modulo(Position);

			switch (m_iUsedSet)
			{
				case 1: m_pvecPositions1[Node] = pos;
						if (is2DRepresentation())
							m_pvecPositions1[Node].z=0;
						;break;
				case 2: m_pvecPositions2[Node] = pos;
						if (is2DRepresentation())
							m_pvecPositions2[Node].z=0;
						;break;
				case 3: m_pvecPositions3[Node] = pos;
						if (is2DRepresentation())
							m_pvecPositions3[Node].z=0;
						;break;
			}

		}
	}

	void SetForce(unsigned int Node, CVector3 Position)
	{
		if (Node<m_uiNumNodes)
		{
			m_pvecForces[Node] = Position;
			if (is2DRepresentation())
				m_pvecForces[Node].z=0;
		}
	}

	void SetIsSelected(unsigned int Node, bool s)	
	{
		if (Node<m_uiNumNodes)
		{
			m_pbSelected[Node] = s;
		}
	}

	
	void AddForce(unsigned int Node, CVector3 force)
	{
		if (Node<m_uiNumNodes)
		{
			m_pvecForces[Node] = m_pvecForces[Node].add(force);
			if (is2DRepresentation())
				m_pvecForces[Node].z=0;
		}
	}

	void AddVelocity(unsigned int Node, CVector3 v)
	{
		if (Node<m_uiNumNodes)
		{
			m_pvecVelocities[Node] = m_pvecVelocities[Node].add(v);
			if (is2DRepresentation())
				m_pvecVelocities[Node].z=0;
		}
	}

		void SetVelocity(unsigned int Node, CVector3 v)
	{
		if (Node<m_uiNumNodes)
		{
			m_pvecVelocities[Node] = v;
			if (is2DRepresentation())
				m_pvecVelocities[Node].z=0;
		}
	}

	/**Add position in current position set*/
	void AddPosition(unsigned int Node, CVector3 pos, cGeometry* geom)
	{
		if (Node<m_uiNumNodes)
		{

			switch (m_iUsedSet)
			{
				case 1: m_pvecPositions1[Node] = m_pvecPositions1[Node].add(pos);
						m_pvecPositions1[Node] = geom->modulo(m_pvecPositions1[Node]);
						if (is2DRepresentation())
							m_pvecPositions1[Node].z=0;
						;break;
				case 2: m_pvecPositions2[Node] = m_pvecPositions2[Node].add(pos);
						m_pvecPositions2[Node] = geom->modulo(m_pvecPositions2[Node]);
						if (is2DRepresentation())
							m_pvecPositions2[Node].z=0;
						;break;
				case 3: m_pvecPositions3[Node] = m_pvecPositions3[Node].add(pos);
						m_pvecPositions3[Node] = geom->modulo(m_pvecPositions3[Node]);
						if (is2DRepresentation())
							m_pvecPositions3[Node].z=0;
						;break;
			}

		}
	}

	void SetVelocitiy(unsigned int Node, CVector3 v)
	{
		if (Node<m_uiNumNodes)
		{
			m_pvecVelocities[Node] = v;
			if (is2DRepresentation())
				m_pvecVelocities[Node].z=0;
		}
	}

	void ClearvecForces()
	{
		for (int i=0; i<m_uiNumNodes; i++)
			m_pvecForces[i].Set(0,0,0) ;
	}

	void ClearvecVelocities()
	{
		for (int i=0; i<m_uiNumNodes; i++)
			m_pvecVelocities[i].Set(0,0,0); 
	}

	void ClearPositions()
	{
		for (int i=0; i<m_uiNumNodes; i++)
			switch (m_iUsedSet)
			{
				case 1: m_pvecPositions1[i].Set(0,0,0) ;break;
				case 2: m_pvecPositions2[i].Set(0,0,0)  ;break;
				case 3: m_pvecPositions3[i].Set(0,0,0) ;break;
			}


	}

	/**Initialize graph*/
	void Init(unsigned int inumnodes)
	{


		m_uiNumNodes = inumnodes;
		m_pbAdjacencmatrix = new bool[m_uiNumNodes*m_uiNumNodes];
		clearConnections();

		m_pvecPositions1 = new CVector3[m_uiNumNodes];
		m_pvecPositions2 = new CVector3[m_uiNumNodes];
		m_pvecPositions3 = new CVector3[m_uiNumNodes];

		m_pvecForces = new CVector3[m_uiNumNodes];;
		m_pvecVelocities = new CVector3[m_uiNumNodes];
		m_pbSelected = new bool[m_uiNumNodes];
		m_psVerticeNames = new string[m_uiNumNodes];
		m_psVerticeLabels = new string[m_uiNumNodes];
		m_psVerticeWeights = new float[m_uiNumNodes];

		for (int i=0; i<m_uiNumNodes;i++)
			m_psVerticeWeights[i]=1;

		m_fMaxScale=8;

		m_iUsedSet=1;
		

	}

	/**Clean up all arrays*/
	void Free()
	{
		free(m_pbAdjacencmatrix);
		free(m_pvecPositions1);
		free(m_pvecPositions2);
		free(m_pvecPositions3);
		free(m_pvecForces);
		free(m_pvecVelocities);
		free(m_pbSelected);
		free(m_psVerticeNames);
		free(m_psVerticeLabels);
		free(m_psVerticeWeights);
	
	}

	cGraph(void)
	{
		m_pbAdjacencmatrix=NULL;
		m_uiNumNodes=0;
		m_fTimeStepMult=1;
		m_fMaxScale=8.0;
	}

	~cGraph(void)
	{
	}



	void clearConnections()
	{
		for (int i=0; i<m_uiNumNodes; i++)
			for (int ii=0; ii<m_uiNumNodes; ii++)
				m_pbAdjacencmatrix[i+m_uiNumNodes*ii]=false;  

		for (int i=0; i<m_uiNumNodes; i++)
			m_pbAdjacencmatrix[i+m_uiNumNodes*i]=true;  
	}

	void ClearIsSelected()
	{
		for (int i=0; i<m_uiNumNodes; i++)
			m_pbSelected[i] = false;

	}
	void setIsConnected(unsigned int Node1, unsigned int Node2, bool connected)
	{
		if ( (Node1<m_uiNumNodes) & (Node2<m_uiNumNodes) )
		{
			m_pbAdjacencmatrix[Node1+m_uiNumNodes*Node2]=connected;  
			m_pbAdjacencmatrix[Node2+m_uiNumNodes*Node1]=connected;  
		}
	}


	bool isConnected(unsigned int Node1, unsigned int Node2)
	{
		if ((Node1<m_uiNumNodes) & (Node2<m_uiNumNodes))
			return m_pbAdjacencmatrix[Node1+m_uiNumNodes*Node2];  
		else
			return false;
	}

	bool isSelected(unsigned int Node1)
	{
		if (Node1<m_uiNumNodes)
			return m_pbSelected[Node1];  
		else
			return false;
		
	}

	void setIs2DRepresentation(bool s)
	{
		m_bRepresentation2D=s;
	}

	bool is2DRepresentation()
	{
		return m_bRepresentation2D;
	}

	void setTimeStepMult(float t)
	{
		m_fTimeStepMult=t;
	}

	float getTimeStepMult()
	{
		return m_fTimeStepMult;
	}

	float getMaxScale() 
	{
		return m_fMaxScale;
	}

	void setUsedPositionSet(int i)
	{
		m_iUsedSet=i;
	}

private:

	/**is current geometry 2D or 3D? If it is 2D the z-value is always kept at zero*/
	bool m_bRepresentation2D;
	/**Number of nodes*/
	unsigned int m_uiNumNodes;
	
	/**Bool-Array containing adjacencematrix*/
	bool *m_pbAdjacencmatrix;

	/**timestep-scale for user-defined simulation-speeds*/
	float m_fTimeStepMult;

	/**Different Position sets. Needed for consistency
	when switching between different view-modes*/
	CVector3 *m_pvecPositions1,*m_pvecPositions2,*m_pvecPositions3;
	/**Current used Position-Set*/
	int m_iUsedSet;

	/**Array holding values of current simulation force-state*/
	CVector3* m_pvecForces;
	/**Array holding values of current simulation velocity-state*/
	CVector3* m_pvecVelocities;

	/**Array holding name-properties of vertices*/
	string* m_psVerticeNames;
	/**Array holding label-properties of vertices*/
	string* m_psVerticeLabels;
	/**Array holding weight-properties of vertices*/
	float* m_psVerticeWeights;
	/**Array holding selected-properties of vertices*/
	bool* m_pbSelected;

	/**Maximum scale encountered in all nodes. Needed to normalize render-size*/
	float m_fMaxScale;

	/**XML Dom-Document*/
	DOMDocument *m_pDoc;
};
