/*SpringEmbedders - A program to visualize graphs
in non-euclidean geometries using a Springembedder Algorithm

Copyright (C) 2005  Markus Lipp

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/


#pragma once

#include "Vector3.h"

/**This class encapsulates the force-calculation routines needed for the springembedders-algorithm*/
class cSpring
{
public:

	/**Set min/max values of parameters*/
	void setSpringParametersBoundaries(float springConstantMins,  float springConstantMaxs,  float springLengthMins,  float springLengthMaxs, float repulsionStrengthMins, float repulsionStrengthMaxs )
	{

		m_fSpringConstantMin=springConstantMins;
		m_fSpringConstantMax=springConstantMaxs;
		m_fSpringLengthMin=springLengthMins;
		m_fSpringLengthMax=springLengthMaxs;
		m_fRepulsionStrengthMin=repulsionStrengthMins;
		m_fRepulsionStrengthMax=repulsionStrengthMaxs;


	}

	/**set repulsion strength relative to min/max boundaries. s=0->min, s=1->max*/
	void setRepulsionStrengthRelative(float s)
	{
		m_fRepulsionStrength= m_fRepulsionStrengthMin+ s*(m_fRepulsionStrengthMax-m_fRepulsionStrengthMin);

	}

	/**set spring parameters relative to min/max boundaries. 0->min, 1->max*/
	void setSpringParametersRelative(float constant, float length)
	{
		m_fSpringConstant= m_fSpringConstantMin + constant*(m_fSpringConstantMax-m_fSpringConstantMin);
		m_fSpringLength= m_fSpringLengthMin + length*(m_fSpringLengthMax-m_fSpringLengthMin);

	}

	/**set spring parameters*/
	void setSpringParameters(float constant, float length)
	{
		m_fSpringConstant=constant;
		m_fSpringLength = length;
	}

	cSpring(void)
	{
		m_fSpringConstant=1;
		m_fSpringLength=1;
		m_fRepulsionStrength=200;
		float m_fRepulsionCutoff=0.05;
	}

	/**Calculate force for connected components. The first vector is always defined as (0,0,0)
	This is due the fact that the tangent-space position (0,0,0) equals the first vector.*/
	CVector3 forceConnected( CVector3 vec2)
	{
		CVector3 vec1 = CVector3(0,0,0);
		CVector3 length = vec1.sub(vec2);
		//spring-force calculation: F=ds*c
		float magnitude = (length.length()-m_fSpringLength) * m_fSpringConstant;
		length.setLength(magnitude);
		return length;
	}

	/**Calculate repulsion strength. The first vector is always defined as (0,0,0)
	This is due the fact that the tangent-space position (0,0,0) equals the first vector.*/
	CVector3 forceRepulsion( CVector3 vec2)
	{
		CVector3 vec1 = CVector3(0,0,0);
		CVector3 length = vec1.sub(vec2);
		float l=length.length();
		if ( !FLOAT_EQ(l,0) )
		{
            //arbitrary force calculation function
			float magnitude = __max(m_fRepulsionStrength/pow(length.length(),2)-m_fRepulsionCutoff,0);
			length.setLength(magnitude);
		}
		return length;
	}


	~cSpring(void)
	{
		
	}

	float getSpringConstant() {return m_fSpringConstant;};
	float getSpringLength() {return m_fSpringLength;};
	float getRepulsionStrength() {return m_fRepulsionStrength;};

	/**This value is subtracted from the calculated repulsion force magnitude. Negative force magnitudes are clamped to zero.*/
	void setRepulsionCutoff(float sfRepulsionCutoff)
	{m_fRepulsionCutoff= sfRepulsionCutoff;};

private:
	float m_fSpringConstant,m_fSpringLength,m_fRepulsionStrength;
	float m_fRepulsionCutoff;

	float m_fSpringConstantMin,  m_fSpringConstantMax,  m_fSpringLengthMin,  m_fSpringLengthMax, m_fRepulsionStrengthMin,
		m_fRepulsionStrengthMax;

};
